from typing import Any, Callable, Dict, List, Tuple

import wrapt
from django.apps.config import AppConfig
from django.contrib.admin.sites import AdminSite
from django.contrib.auth.decorators import user_passes_test
from django.urls import URLPattern, path, reverse

from django_managerie.blocklist import COMMAND_BLOCKLIST
from django_managerie.commands import ManagementCommand, get_commands
from django_managerie.types import CommandMap
from django_managerie.views import ManagerieCommandView, ManagerieListView

superuser_required = user_passes_test(lambda u: u.is_active and u.is_superuser)


class Managerie:
    ignored_app_names = {
        'django.core',
        'django.contrib.staticfiles',
    }

    def __init__(self, admin_site: AdminSite) -> None:
        self.admin_site = admin_site

    def patch(self) -> None:
        if hasattr(self.admin_site, 'patched_by_managerie'):
            return
        old_get_app_list = self.admin_site.get_app_list
        self.admin_site.get_app_list = self.patched_get_app_list(old_get_app_list)  # type: ignore[assignment]
        old_get_urls = self.admin_site.get_urls
        self.admin_site.get_urls = self.patched_get_urls(old_get_urls)  # type: ignore[assignment]
        self.admin_site.patched_by_managerie = True  # type: ignore[attr-defined]

    def is_command_allowed(self, command: ManagementCommand) -> bool:
        # This could be overridden in a subclass to allow for more fine-grained
        # control over which commands are allowed.

        if command.full_name in COMMAND_BLOCKLIST:
            return False
        if getattr(command.get_command_class(), 'disable_managerie', False):
            return False
        return True

    def get_commands(self) -> Dict[AppConfig, CommandMap]:
        command_map: Dict[AppConfig, CommandMap] = {}
        for app_config, commands in get_commands().items():
            command_map[app_config] = {
                command_name: command
                for (command_name, command)
                in commands.items()
                if self.is_command_allowed(command)
            }
        return command_map

    def get_commands_for_app_label(self, app_label: str) -> CommandMap:
        for app_config, commands in self.get_commands().items():
            if app_config.label == app_label:
                return commands
        return {}

    @wrapt.decorator
    def patched_get_app_list(self, wrapped, instance: AdminSite, args, kwargs):
        request = args[0]
        app_list = wrapped(*args, **kwargs)
        if request.user.is_superuser:
            self._augment_app_list(app_list)
        return app_list

    @wrapt.decorator
    def patched_get_urls(self, wrapped: Callable, instance: AdminSite, args: Tuple[()], kwargs: Dict[Any, Any]) -> list:
        urls = wrapped(*args, **kwargs)
        return self._get_urls() + list(urls)

    def _augment_app_list(self, app_list):
        all_commands = {app_config.label: commands for (app_config, commands) in self.get_commands().items()}
        for app in app_list:
            commands = all_commands.get(app['app_label'], [])
            if commands:
                app.setdefault('models', []).append({
                    'perms': {'change': True},
                    'admin_url': reverse(
                        'admin:managerie_list',
                        kwargs={'app_label': app['app_label']},
                        current_app=self.admin_site.name,
                    ),
                    'name': 'Commands',
                    'object_name': '_ManagerieCommands_',
                })

    def _get_urls(self) -> List[URLPattern]:
        return [
            path(
                'managerie/<app_label>/<command>/',
                superuser_required(ManagerieCommandView.as_view(managerie=self)),
                name='managerie_command',
            ),
            path(
                'managerie/<app_label>/',
                superuser_required(ManagerieListView.as_view(managerie=self)),
                name='managerie_list',
            ),
            path(
                'managerie/',
                superuser_required(ManagerieListView.as_view(managerie=self)),
                name='managerie_list_all',
            ),
        ]

    @property
    def urls(self):
        return (self._get_urls(), 'admin', self.admin_site.name)



import argparse
import warnings
from typing import Any, Mapping, Optional, Type

from django import forms

BOOLEAN_ACTIONS = (argparse._StoreTrueAction, argparse._StoreFalseAction, argparse._StoreConstAction)

FIELD_CLASS_MAP: Mapping[Any, Type[forms.Field]] = {
    float: forms.FloatField,
    int: forms.IntegerField,
    None: forms.CharField,
}


class ArgumentParserForm(forms.Form):
    IGNORED_DESTS = {
        'force_color',
        'no_color',
        'pythonpath',
        'settings',
        'traceback',
    }

    def __init__(self, *, parser: argparse.ArgumentParser, **kwargs) -> None:
        super().__init__(**kwargs)
        self.parser = parser
        for action in self.parser._actions:
            self._process_action(action)

    def _process_action(self, action: argparse.Action) -> None:
        if isinstance(action, argparse._HelpAction):
            return
        if action.dest in self.IGNORED_DESTS:
            return
        field_cls: Optional[Type[forms.Field]] = None
        field_kwargs = dict(
            initial=action.default,
            label=action.dest.replace('_', ' ').title(),
            help_text=action.help,
            required=action.required,
        )
        if isinstance(action, BOOLEAN_ACTIONS):
            field_cls = forms.BooleanField
        elif isinstance(action, argparse._StoreAction):
            if action.type not in FIELD_CLASS_MAP:
                warnings.warn(f'No specific field class for type {action.type!r}')
            field_cls = FIELD_CLASS_MAP.get(action.type, forms.Field)
        if field_cls:
            self.fields[action.dest] = field_cls(**field_kwargs)

        # TODO: Probably support for more fields :)



import io
import time
import traceback
from contextlib import redirect_stderr, redirect_stdout
from itertools import chain
from typing import TYPE_CHECKING, Any, Dict, Optional

from django.apps import apps
from django.apps.config import AppConfig
from django.http import Http404, HttpResponse
from django.views.generic import FormView, TemplateView

from django_managerie.commands import ManagementCommand
from django_managerie.forms import ArgumentParserForm

if TYPE_CHECKING:
    from django_managerie.managerie import Managerie


class MenagerieBaseMixin:
    managerie: Optional["Managerie"] = None
    kwargs: Dict[str, Any]
    _app: Optional[AppConfig]

    def get_app(self) -> Optional[AppConfig]:
        if hasattr(self, '_app'):
            return self._app
        if 'app_label' in self.kwargs:
            self._app = apps.get_app_config(self.kwargs['app_label'])
            return self._app
        return None


class ManagerieListView(MenagerieBaseMixin, TemplateView):
    template_name = 'django_managerie/admin/list.html'

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context['app'] = app = self.get_app()
        context['title'] = f"{app.verbose_name if app else 'All Apps'} – Commands"
        managerie = self.managerie
        assert managerie
        context['commands'] = sorted(
            (
                managerie.get_commands_for_app_label(app.label).values()
                if app
                else chain(*(app_commands.values() for app_commands in managerie.get_commands().values()))
            ),
            key=lambda cmd: cmd.full_title
        )
        return context


class ManagerieCommandView(MenagerieBaseMixin, FormView):
    template_name = 'django_managerie/admin/command.html'

    @property
    def command_name(self) -> str:
        return self.kwargs['command']

    def get_command_object(self) -> ManagementCommand:
        app = self.get_app()
        managerie = self.managerie
        assert app and managerie
        try:
            return managerie.get_commands_for_app_label(app.label)[self.command_name]
        except KeyError:
            raise Http404(f"Command {self.command_name} not found in {app.label}")

    def get_form(self, form_class=None) -> ArgumentParserForm:
        cmd = self.get_command_object().get_command_instance()
        parser = cmd.create_parser('django', self.command_name)
        return ArgumentParserForm(parser=parser, **self.get_form_kwargs())

    def get_context_data(self, **kwargs) -> Dict[str, Any]:
        context = super().get_context_data(**kwargs)
        command = self.get_command_object()
        context.update(
            app=self.get_app(),
            command=command,
            command_help=command.get_command_instance().help,
            title=command.full_title,
        )
        return context

    def form_valid(self, form: ArgumentParserForm) -> HttpResponse:
        # This mimics BaseCommand.run_from_argv():
        options = dict(form.cleaned_data)
        args = options.pop('args', ())  # "Move positional args out of options to mimic legacy optparse"
        stdout = io.StringIO()
        stderr = io.StringIO()
        error = None
        error_tb = None
        t0 = time.time()
        co = self.get_command_object()
        with redirect_stdout(stdout), redirect_stderr(stderr):
            options.update({
                'traceback': True,
                'no_color': True,
                'force_color': False,
                'stdout': stdout,
                'stderr': stderr,
            })
            cmd = co.get_command_instance()
            try:
                cmd._managerie_request = self.request
                cmd.execute(*args, **options)
            except SystemExit as se:  # We don't want any stray sys.exit()s to quit the app server
                stderr.write(f'<exit: {se}>')
            except Exception as exc:
                error = exc
                error_tb = traceback.format_exc()
        context = self.get_context_data(
            executed=True,
            form=form,
            error=error,
            error_tb=error_tb,
            stdout=stdout.getvalue(),
            stderr=stderr.getvalue(),
            duration=(time.time() - t0),
        )
        return self.render_to_response(context=context, status=(400 if error else 200))



import os
from collections import OrderedDict, defaultdict
from functools import lru_cache
from importlib import import_module

from django.apps import apps
from django.core.management import find_commands
from django.urls import reverse


class ManagementCommand:
    def __init__(self, app_config, name):
        self.app_config = app_config
        self.name = name
        self.title = self.name.replace('_', ' ').title()

    @property
    def url(self):
        return reverse(
            'admin:managerie_command',
            kwargs={'app_label': self.app_config.label, 'command': self.name},
        )

    def get_command_class(self):
        mname = f'{self.app_config.name}.management.commands.{self.name}'
        return import_module(mname).Command

    def get_command_instance(self):
        cls = self.get_command_class()
        return cls()

    @property
    def full_title(self):
        return f'{self.app_config.verbose_name} – {self.title}'

    @property
    def full_name(self):
        return f'{self.app_config.label}.{self.name}'


@lru_cache(maxsize=None)
def get_commands():
    # Logic filched from django.core.management.get_commands(), but expressed in a saner way.
    apps_to_commands = defaultdict(OrderedDict)

    for app_config in apps.get_app_configs():
        path = os.path.join(app_config.path, 'management')
        for command_name in find_commands(path):
            apps_to_commands[app_config][command_name] = ManagementCommand(
                app_config=app_config,
                name=command_name,
            )
    return apps_to_commands
